"""Automatic generation of test sets for OddOneOut and Topk methods using Wikidata"""
import requests
import sys
from SPARQLWrapper import SPARQLWrapper, JSON
from wikidata.client import Client
import re

def construct_query(item, language):
    """Constructs a basic SPARQL query using the 'instance of' property.

    Parameters
    ----------
    item : str 
        Wikidata item
    lang : str
        two-letter wikimedia language code

    Returns
    -------
    str
        the new language appropriate SPARQL query
    str
        a category title corresponding to the query

    """
    # use the 'instance of' property to construct categories
    instance_of = 'P31'
    #Wikidata client
    client = Client()
    entity_label = str(client.get(item).label)
    title = 'instance of '+entity_label   
    query = """#"""+title+"""
        SELECT DISTINCT ?item ?label
        WHERE
        {
        ?item wdt:"""+instance_of+""" wd:"""+item+""".
        # Only keep single words - move back into query if needed
        filter(!contains(?label," "))
        # Language
        ?item rdfs:label ?label filter(lang(?label) = '"""+language+"""').
        }
        ORDER BY ?label"""

    # queries.update({title:query})
    return query, title

# result = construct_query(['Q1369421'])
# print(result)

# returns results of the SPARQL query
def get_results(query, endpoint_url):
    """makes request to Wikidata SPARQL endpoint and returns result of query.

    Parameters
    ----------
    query : str 
        SPARQL query
    endpoint_url : str
        the url for the appropriate SPARQL endpoint

    Returns
    -------
    dict
        the results of the SPARQL query

    """
    user_agent = "WDQS-example Python/%s.%s" % (sys.version_info[0], sys.version_info[1])
    # TODO adjust user agent; see https://w.wiki/CX6
    sparql = SPARQLWrapper(endpoint_url, agent=user_agent)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    return sparql.query().convert()

# format SPARQL results for a single category (a single query)
def get_category(query, endpoint_url="https://query.wikidata.org/sparql"):
    """query wikidata to get the list of words belonging to a category

    Parameters
    ----------
    query : str 
        the SPARQL query representing the category to be constructed
    endpoint_url : str, optional
        the appropriate SPARQL endpoint (default Wikidata endpoint)

    Returns
    -------
    list
        a list of words which make up the desired category

    """
    # form the query
    # query = construct_query(item)[0]
    results_dict = get_results(query, endpoint_url)
    values = []
    for result in results_dict["results"]["bindings"]:
        item = result['label']['value'].lower()
        # handle sparql queries that return synonyms '/' separated
        if re.search('/',item):
            # split into two words and append both
            split = item.split('/')
            values.append(split[0])
            values.append(split[1])
        else:
            values.append(item)        
    return values


# create a separate test set for each language from queries.
# test set formatted to be compatible with OddOneOut and Topk
def generate_test_set(items, languages, filename, lemmatized=False, all=False):
    """Generate a complete test set for use with OddOneOut and Topk

    Parameters
    ----------
    query : list of str 
        list of wikidata categories to be included in the test set
    languages : list of str
        list of all wikimedia languages to translate test set into
    filename : str
        location to save test sets

    Returns
    -------
    None
        .txt files will be generated and saved at location specified in filename parameter
    """
    for lang in languages:
        with open(filename+'_'+lang+'.txt','w') as f:
            for i in items:
                # creates a query
                query_obj = construct_query(i,lang)
                query = query_obj[0]
                title = query_obj[1]
                # Get the items in the category formed by current query
                values = get_category(query)
                # print('query=',query)
                # print('title=',title)
                # print('values=',values)
                f.write(':'+title+' ('+lang+')')
                f.write('\n')
                f.writelines('%s ' % item for item in values)
                f.write('\n')



"""Implementation of the Topk and OddOneOut methods for evaluating word embeddings"""

# dependencies
from gensim.models import Word2Vec
import re
from itertools import combinations
import random
import warnings

def Topk(cat_file, model, k=3, allow_oov=False, ft=False):
    """Performs Topk Evaluation on an embedding model

    Parameters
    ----------
    cat_file : str 
        location of the .txt file of categories to evaluate on
    model : 
        embedding model loaded via gensim
    k : int 
        number of words to return from similarity query
    allow_oov : bool, optional
        whether to skip or mark wrong comparisons with oov words (default False)
    ft : bool, optional
        set to True if using fastText embedding models (default False)

    Returns
    -------
    float
        overall accuracy
    dict
        category accuracy (float accuracy for each category)
    list
        list of skipped categories
    int
        overall raw score (total number of correct comparisons)
    dict
        category raw score (number of correct comparisons for each category)

    """
    cats = {}
    skipped_cats = []
    # convert test set file into dictionary where each category is a key
    # and the corresponding value is the list of words belonging to it.
    with open(cat_file, 'r') as f:
        for line in f:
            if re.search('^:',line):
                key = line.rstrip()
            # can't do topk comparisons when category has fewer than 2 words
            elif not len(line.strip().split()) < 2:
                cats.update({key: line.strip().split()})
            else:
                skipped_cats.append(key)

    model_vocab = model.wv.vocab
    # fast text uses n-grams, so checking for oov is different
    if ft:
        model_vocab = model.wv
        print('using ft n-grams')
    oov_list = []
    words_in_test = 0
    # verify that all words are in the vocabulary
    for value in cats.values():
        for word in value:
            words_in_test += 1
            # keep track of oov words
            if allow_oov and word not in model_vocab:
                oov_list.append(word)
            elif not allow_oov and word not in model_vocab:
                raise KeyError('word '+word+' is not in vocabulary')
    
    # test set info
    print('Performing Topk Evaluataion')
    print(' %d categories do not have enough words and will be skipped' % len(skipped_cats))
    print(' %d words have been identified as out out of vocabulary' % len(oov_list))
    print(' %d total words in test set' % words_in_test)
    ratio = len(oov_list)/words_in_test
    print('out of vocab ratio is ','{0:.2f}'.format(ratio))

    # for tracking scores for each category
    category_acc = {}
    category_raw = {}
    # total raw number of correct answers
    raw_correct = 0
    # total number of categories
    m = len(cats.items())

    # find the topk most similar for each word in all categories
    for cat, words in cats.items():
        # zero out for each new category
        cat_score = 0

        for word in words:
            # number of the words 
            n = len(words)
            category = words
            ignore_comparison=False
            # if word isn't in vocab, then ignore comparison altogether
            if word in oov_list:
                ignore_comparison = True
            # Evaluating Comparisons
            if not ignore_comparison:
                # find top_k similar words for a given entry
                topk = model.wv.most_similar(positive=word, topn=k)
                # items in each category
                for x in topk:
                    # x is a tuple and we want first element
                    cat_score += x[0] in category

        # Update category score
        category_acc.update({cat: cat_score/(n*k)})
        category_raw.update({cat: cat_score})
        # update total raw number correct
        raw_correct += cat_score
    
    # Total Score
    accuracy = sum(category_acc.values())/m

    return accuracy, category_acc, skipped_cats, raw_correct, category_raw


def OddOneOut(cat_file, model, k_in, sample_size=1000, restrict_vocab=None, allow_oov=False, ft=False, debug=False):
    """Performs Topk Evaluation on an embedding model

    Parameters
    ----------
    cat_file : str 
        location of the .txt file of categories to evaluate on
    model : 
        embedding model loaded via gensim
    k_in : int 
        size of group of words from the same category
    sample_size : int
        number of OddOneOut comparisons to evaluate for each category (default=1000)
    restrict_vocab : int, optional
        the size of the model vocabulary to sample the out-word from (default None)
    allow_oov : bool, optional
        whether to skip or mark wrong comparisons with oov words (default False)
    ft : bool, optional
        set to True if using fastText embedding models (default False)
    debug : bool, optional
        print out the comparisons being made for debugging purposes (default False)

    Returns
    -------
    float
        overall accuracy
    dict
        category accuracy (float accuracy for each category)
    list
        list of skipped categories
    int
        overall raw score (total number of correct comparisons)
    dict
        category raw score (number of correct comparisons for each category)

    """
    original_state = random.getstate()
    random.seed(0)
    cats = {}
    skipped_cats = []
    # create category dictionary with key as category and value = list of words from that category
    with open(cat_file, 'r') as f:
        for line in f:
            if re.search('^:',line):
                key = line.rstrip()
            # check category contains enough values
            elif len(line.strip().split()) < k_in:
                skipped_cats.append(key)
            else:
                cats.update({key: line.split()})
  
    # Verify all words are in the vocabulary
    model_vocab = model.wv.vocab
    # fast text uses n-grams, so checking for oov is different
    if ft:
        model_vocab = model.wv
        print('using ft n-grams')
    oov_list = []
    words_in_test = 0
    for value in cats.values():
        for word in value:
            words_in_test += 1
            if allow_oov and word not in model_vocab:
                oov_list.append(word)
            elif not allow_oov and word not in model_vocab:
                raise KeyError('word '+word+' is not in vocabulary')
    
    # test set info
    print('OddOneOut Evaluation')
    print(' %d categories have fewer than k_in entries and will be skipped' % len(skipped_cats))
    print(' %d words have been identified as out out of vocabulary' % len(oov_list))
    print('words_in_test=',words_in_test)    
    ratio = len(oov_list)/words_in_test
    print('out of vocab ratio is ','{0:.2f}'.format(ratio))


    # for storing accuracies
    category_acc = {}
    category_raw = {}
    
    # to return raw # correct preds instead of accuracy
    raw_correct = 0

    # total number of categories
    m = len(cats.items())
    print('Will calculate the %d th order OddOneOut score for %d categories' %(k_in,m))
    not_skipped = 0
    for cat in cats.keys():     
        s = list(combinations(cats[cat],k_in))
        c_i = cats[cat]
        # sample k-combos
        s_sampled = random.choices(s, k=sample_size)
        # print(s_sampled)

        # sample OddOneOut from model vocabulary
        w_sampled = []
        while len(w_sampled) < sample_size:
            word = random.choice(model.wv.index2word[:restrict_vocab])
            # don't add word if it's a dupe
            if word not in c_i:
                w_sampled.append(word)
        # kth order OddOneOut score for category i
        cat_score = 0
        # compute OddOneOut for each comparison
        for in_words, odd_one_out in zip(s_sampled,w_sampled):
            comparison = in_words + (odd_one_out,)

            # By default don't ignore any comparisons
            ignore_comparison = False
            if allow_oov:
                # check for oov word
                for w in comparison:
                    if w in oov_list:
                        ignore_comparison = True
            if not ignore_comparison:
                cat_score += int(model.wv.doesnt_match(comparison)==odd_one_out)
                if debug:
                    not_skipped +=1
                    print('comparison count=',not_skipped)
                    print('comparison=',comparison)
                    print('predicted OddOneOut=',model.wv.doesnt_match(comparison))
                    print('actual OddOneOut=',odd_one_out)
                # print('cat_score=',cat_score)
        category_acc.update({cat: cat_score/sample_size})
        category_raw.update({cat:cat_score})
        # update total raw number correct answers
        raw_correct += cat_score
    # Calculate Total Score
    accuracy = sum(category_acc.values())/m
    random.setstate(original_state)

    return accuracy, category_acc, skipped_cats, raw_correct, category_raw



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="gensim-evaluations",
    version="0.0.4",
    description="Methods for evaluating low-resource word embedding models trained with gensim",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/n8stringham/gensim-evaluations",
    author="Nate Stringham",
    author_email="n8stringham@gmail.com",
    license="LGPL-2.1",
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["gensim_evaluations"],
    include_package_data=True,
    install_requires=["requests", "sparqlwrapper","Wikidata","gensim",], 
    python_requires=">=3.7",

)